/*
 * errdb.c
 *
 * Part of the strings db toolkit.
 *
 * This is the implementation for the functions prototyped in errdb.h.
 * It's part as a miniature library for the toolkit.
 *
 * 0.4.3: implemented
 *
 * Copyright (c) 1996 Amy A. Lewis
 * Distributed under GPL.  See COPYING.
 *
 * $Id: errdb.c 13 2004-03-27 20:21:34Z amyzing $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include "strdb.h"
#include "keydb.h"	/* for sdb_unicorn */
#include "errdb.h"	/* ditto */

const char *sdb_err_string[] = {
    "\n",
    "Not a strdb index file (bad magic)\n",
    "Not a strdb key file (bad magic)\n",
    "Index file outdated\n",
    "Key file outdated\n",
    "Index out of range\n",
    "Usage\n",
    "Incorrect major version\n",
    "File read error\n",
    "File write error\n",
    "File not found\n",
    "\n",
    "Memory allocation error\n",
    "\n",
    "\n",
    "\n",
    "Total confusion.  SNAFU.  Shoot the programmer.\n"
};

/*
 * sdb_unicorn
 * (The name is from a song I learned in college.  Refrain is "And magic is
 * there,/and magic will stay,/as long as the unicorn passes your way.")
 *
 * Checks a file thought to be an index file, compares the magic in the
 * file with the constant sdb_MAGIC.
 * On success, returns 0.  Counterintuitive, but we have lots of ways to fail.
 * On "normal" failure of the magic, returns 1.
 * If the file is outdated, returns 2.
 * On file error (can't find one, it's too small) returns 3.
 * On file read error, returns 4.
 * Returns 5 for memory allocation error.
 */
int 
sdb_unicorn(FILE *idxf, char ftype)
{
    sdb_header *sdbh;
    sdb_key_head *sdbkh;
    struct stat datastat;
    
    if (!(idxf))
        return sdb_ENOENT;
    if (fstat(fileno(idxf), &datastat) == -1)
        return sdb_ENOENT;

    if (!ftype)		/* index file */
    {
        if (datastat.st_size < sizeof(sdb_header))
            return sdb_ENOTIDX;	/* too small to be an index file */
        else
        {
            rewind(idxf);
            sdbh = (sdb_header *)malloc(sizeof(sdb_header));
            if (!sdbh)
                return sdb_EMEMFAULT;		/* memory allocation failure */
            if (!fread(sdbh, sizeof(sdb_header), 1, idxf))
                return sdb_EFREAD;		/* screwed something up */
            if (strcmp(sdb_MAGIC, sdbh->sdb_magic) != 0)
                return sdb_ENOTIDX;
            if (stat(sdbh->sdb_filename, &datastat) == -1)
                return sdb_ENOENT;
            else
            {
                if (datastat.st_mtime != sdbh->sdb_mtime)
                    return sdb_EIDXDATE;
            }
            return 0;
        }
    }
    else		/* key file */
    {
        if (datastat.st_size < sizeof(sdb_key_head))
            return sdb_ENOTKEY;		/* too small to be a key file */
        else
        {
            rewind(idxf);
            sdbkh = (sdb_key_head *)malloc(sizeof(sdb_key_head));
            if (!sdbkh)
                return sdb_EMEMFAULT;		/* memory allocation failure */
            if (!fread(sdbkh, sizeof(sdb_key_head), 1, idxf))
                return sdb_EFREAD;		/* screwup */
            if (strcmp(sdb_MAGIC, sdbkh->sdb_magic) != 0)
                return sdb_ENOTKEY;
            if (stat(sdbkh->sdb_indexfile, &datastat) == -1)
                return sdb_ENOENT;
            else
            {
                if (datastat.st_mtime != sdbkh->sdb_mtime)
                    return sdb_EKEYDATE;
            }
            return 0;		/* passed all tests */
        }
    }
    /* if you get here, there's a lot of damned trouble */
    return sdb_ECHAOS;
}
