/*
 * maintdb.c
 *
 * Part of the strings db package.  maintdb.h contains prototypes for a few
 * selected common record maintenance tasks.  These primitives may not suit
 * the purposes of a particular application.
 *
 * version 0.4.1, April 13, 1997: first pass.
 *
 * Copyright (c) 1997 Amelia A. Lewis
 * Distributed under GPL.  See COPYING.
 *
 * $Id: maintdb.c 13 2004-03-27 20:21:34Z amyzing $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <ctype.h>
#include "strdb.h"
#include "maintdb.h"

const char *sdb_maint_err[] =
{
    "No data.\n",
    "Record terminates before beginning.\n",
    "Record does not end with terminator.\n",
    "Line too long.\n",
    "Record contains internal terminator.\n",
    "Format error.\n",
    "Unknown error.\n"
};

/*
 * sdb_deleterec
 *	Deletes the record indicated by the offset given in the second
 * parameter, in the file indicated by the stream passed as the first
 * parameter.  "Delete" means to mark deleted, to change the terminator
 * string to the special string indicating that the record has been deleted.
 * The record is still indexed until the index file is regenerated.
 *
 * Returns 1 on success, 0 on error.
 */
int
sdb_deleterec(FILE *d, off_t o, unsigned short siz)
{
    char delim[6];
    char sigc;
    int oset;
    char delmark = '\b';

    if (!d)
        return 0;
    
    if (fseek(d, o + siz - 4, SEEK_SET))	/* returns 0 on success */
        return 0;
    
    fread((char *)delim, sizeof(char), 4, d);
    if (strchr(delim, '\n'))
    {
        oset = strchr(delim, '\n') - delim;
    }
    else
        oset = 0;
    if (isgraph(delim[oset + 1]))
        sigc = delim[oset + 1];
    else /* damn.  Get it from previous rec, if possible */
    {
        if (fseek(d, o - 2, SEEK_SET))
            return 0;
        else
            fread((char *)&sigc, sizeof(char), 1, d);
    }

    fseek(d, o + siz - 4, SEEK_SET);
    sprintf(delim, "\n%c%c\n", sigc, delmark);
    if (!fwrite((char *)&delim, sizeof(char), 4, d))
        return 0;
    return 1;
}

/*
 * sdb_purgerec
 *	Bypasses the need to regenerate an index file for a deleted record.
 * It accomplishes this by copying all the pointers in the index file from
 * the one right after the one to be deleted onward.  It then truncates the
 * file at the position of the purged pointer, writes out the saved pointers,
 * then writes the total number of offsets value, decremented by one, back
 * to the file.
 *	Before it does its index file magic, it reads the pointer in, finds
 * the data file, copies it into a temporary file or buffer, truncates the
 * original just before the start of the record to be purged, seeks in
 * the copy to the position immediately after the record to be purged,
 * and writes out the remainder to the previously-truncated original file.
 *	Finally, times are necessarily updated.
 *
 * Note: the damned file *better* have been opened read-write.  If not,
 * we can't do anything with it, because we can't get the file name from
 * the stream, by any means I know.
 *
 * Returns total number of indices remaining on success, 0 on error.
 */
int
sdb_purgerec(FILE *idxf, unsigned long which)
{
    FILE *df;
    sdb_recptr *srp;
    sdb_header sdbh;
    char *bigbuf;
    struct stat filestats;
    int bsize, xsize;
    
    if (!idxf)
        return 0;
    rewind(idxf);
    if (!fread((sdb_header *)&sdbh, sizeof(sdb_header), 1, idxf))
        return 0;
    srp = sdb_pointer(idxf, which);
    if (!srp)
        return 0;
    df = fopen(sdbh.sdb_filename, "r+");
    if (!df)
        return 0;

    if (stat(sdbh.sdb_filename, &filestats) == -1)
        return 0;
    /* read the remainder of the file in, from after the record to be deleted */
    bigbuf = (char *)calloc(filestats.st_size - (srp->data_os + srp->data_size) + 1, sizeof(char));
    if (!bigbuf)
        return 0;
    if (fseek(df, srp->data_os + srp->data_size + 1, SEEK_SET))
        return 0;
    bsize = fread((char *)bigbuf, sizeof(char), filestats.st_size - (srp->data_os + srp->data_size), df);
    if (ftruncate(fileno(df), srp->data_os) != 0)
        return 0;

    fseek(df, 0L, SEEK_END);
    fwrite((char *)bigbuf, sizeof(char), bsize, df);
    fclose(df);
    free(bigbuf);
    bigbuf = NULL;
    
    /* that means we've finished screwing up the data file.  Now wreck the
       index file */
    if (fstat(fileno(idxf), &filestats) == -1)
        return 0;
    xsize = filestats.st_size - (sizeof(sdb_header) + ((sizeof(off_t) + sizeof(unsigned short)) * which)) + 1;
    bigbuf = (char *)calloc(xsize, sizeof(char));
    xsize--;
    if (!bigbuf)
        return 0;
    fseek(idxf, sizeof(sdb_header) + ((sizeof(off_t) + sizeof(unsigned short)) * which), SEEK_SET);
    bsize = fread((char *)bigbuf, sizeof(char), xsize, idxf);

    ftruncate(fileno(idxf), sizeof(sdb_header) + ((sizeof(off_t) + sizeof(unsigned short)) * (which - 1)));
    fseek(idxf, 0L, SEEK_END);
    fwrite((char *)bigbuf, sizeof(char), bsize, idxf);
    
    /* fix the header values and write it out */
    sdbh.sdb_total--;
    stat(sdbh.sdb_filename, &filestats);
    sdbh.sdb_mtime = filestats.st_mtime;
    rewind(idxf);
    fwrite((sdb_header *)&sdbh, sizeof(sdb_header), 1, idxf);
    return 1;
}

/*
 * sdb_parserec
 *	Evaluates a record, passed as a null-delimited string in the first
 * parameter, for correctness.  That is, the record may not contain a term-
 * inator, except in the appropriate position, where it must appear.  Lines
 * beginning with the signal character (passed as the second parameter)
 * which do not qualify as delimiters are tagged as errors.  Lines exceeding
 * the value of MAXLINELEN in length are tagged as errors.  If parameter
 * three is not NULL, the record is checked (roughly) for formatting: the
 * ambiguous field always appears first, and the character string indicates
 * the order that other fields must appear, if the first character in the
 * string is space, or the fields that must appear if the first character
 * is anything other than a space.  The string otherwise contains valid
 * field delimiter characters.  If the character is followed by the signal
 * character, it is repeatable.
 *
 * Errors are returns in the linerr structure, which is allocated here.  If
 * no error is found, the function returns NULL.
 */
linerr*
sdb_parserec(char *rd, char sc, char *format)
{
    linerr *serror = NULL;
    char curline = 0;
    char *aline = NULL,
         *linend = NULL,
         newline[MAXLINELEN];

    serror = (linerr *)malloc(sizeof(linerr));
    serror->lineno = 0;
    serror->errcode = 0;
    
    if (!rd)
    {
        serror->errcode = sdb_EMNODATA;
        return(serror);
    }
    
    /* check that the record doesn't begin with a terminator */
    if (sdb_delim(rd, sc) == 1)
    {
        serror->errcode = sdb_EMEARLYTERM;
        return(serror);
    }
    /* check that the record does end with a terminator */
    aline = strrchr(rd, '\n');
    if (aline && (sdb_delim(aline, sc) == 1))
    {
        serror->errcode = sdb_EMNOTERM;
        return(serror);
    }
    aline = rd;
    do
    {
        curline++;
        linend = strchr(aline, '\n');
        if (*newline)
            *newline = '\0';
        if (linend && (linend > aline))
        {
            if (linend - aline + 1 > 1023)
            {
                /* the line is too long.  break */
                serror->lineno = curline;
                serror->errcode = sdb_EMLONGLINE;
                return(serror);
            }
            strncpy(newline, aline, linend - aline + 1);
            newline[linend - aline + 1] = '\0';
        }
        /* the line is a terminator break */
        if (sdb_delim(newline, sc) == 1)
        {
            serror->lineno = curline;
            serror->errcode = sdb_EMINTERNTERM;
            return(serror);
        }
        if (format)
        {
            /* line looks like a delimiter, but isn't.  break */
            /* line is a delimiter */
                /* we're checking for ordering, and this is out of order. break */
                /* it's in order.  delete it from the string passed in if not multi */
                /* we're checking for formatting, and it's not in the list. break */
                /* it's in the list.  delete it (if not multi). */
        }
        aline = linend + 1;
    } while (*newline);
    return NULL;
}

/*
 * Check the end of an open data file.  It should have a record terminator
 * as the final characters.  If it doesn't, return 0.
 * file and signal character are passed in.
 */
int
sdb_check_df(FILE *df, char sc)
{
    off_t curpos;
    char terminator[5], fileterm[5];
    
    if (!df || !(isgraph(sc)))
        return 0;
    curpos = ftell(df);
    sprintf(terminator, "\n%c%c\n", sc, sc);
    fileterm[4] = '\0';
    
    fseek(df, -4, SEEK_END);
    fread((char *)fileterm, sizeof(char), 4, df);
    fseek(df, curpos, SEEK_SET);
    
    return(!(strcmp(terminator, fileterm)));
}

int
sdb_addterm(sdb_recptr *rec, char sigc, int n)
{
    char *tempdata;
    char catstr[8];
    int addlen = 1;
    
    if (!rec || !rec->data)
        return 0;
    sdb_stripterm(rec, sigc);	/* why take chances? */
    if (rec->data[rec->data_size - 1] == '\n')
    {
        addlen += n;
        (n == 1) ? sprintf(catstr, "%c\n", sigc) : sprintf(catstr, "%c%c\n", sigc, sigc);
    }
    else
    {
        addlen += n + 1;
        (n == 1) ? sprintf(catstr, "\n%c\n", sigc) : sprintf(catstr, "\n%c%c\n", sigc, sigc);
    }
    tempdata = (char *)calloc(rec->data_size + addlen + 1, sizeof(char));
    strcpy(tempdata, rec->data);
    strcat(tempdata, catstr);
    free(rec->data);
    rec->data = tempdata;
    rec->data_size = strlen(rec->data);
    return addlen;
}
