/*
 * cyclestr.c
 *
 * Part of the strings db toolkit.
 *
 * cyclestr is the most ambitious of the core utilities, because it is
 * interactive (unlike all the others).  It is chiefly intended for use in
 * maintenance of the database, though it is not user-friendly and won't
 * particularly guarantee that the database will end in a usable state
 * (however, running idxstr on the data file afterwards should fix things;
 * the data file is not very complex, after all).
 *
 * It opens the data file pointed to by the index file given it, and displays
 * the first record (in raw form).  If the record scrolls off the screen,
 * tough.  If you don't have a scrollback buffer, you're out of luck.  On
 * a line below the record, it displays its prompt:
 * d,n,p,q,w>
 * The letters are the available commands:
 * d: delete this record (mark it deleted, that is--it won't get indexed,
 *    but the text will be there until/unless you use dumpstr to clean
 *    up the data file)
 * n: go to the next record (display it)
 * p: go to the previous record (display it)
 * q: quit
 * w: write this record to a file.  It will prompt for a filename:
 * filename?>
 * If the file exists, it appends (no choice to overwrite).
 *
 * Arrow keys and such are not implemented, and won't be.
 *
 * It does wrap from first record to last, and from last record to first.
 *
 * Look, it isn't *supposed* to be pretty, it's supposed to be *functional*.
 * Use it, for example, in combination with findstr to normalize author's
 * names.  There are a lot of other unix tools that operate on text files,
 * too, so treat this one as just a shortcut to vgrepping records.
 *
 * We *may* eventually add edit as an option, but then we either generate
 * a new index file, which is usually not such a good idea to do automatically
 * (because idxstr has bunches of options for sorting, and doesn't store all
 * of them in the index file), or accept that if the record length is
 * changed by editing (almost a given, really) the remainder of the cycling
 * will show some form of record truncation (much worse if the record is
 * shortened, as a rule, because of the way sdb_record finds the end of
 * a record).
 *
 * version 0.2.1, grotty, December 4, 1996
 *
 * Copyright (c) 1996 Amelia A. Lewis
 * Distributed under GPL.  See COPYING.
 *
 * $Id: cyclestr.c 13 2004-03-27 20:21:34Z amyzing $
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include "../lib/strdb.h"
#include "../lib/errdb.h"
#include "../lib/maintdb.h"

char index_fn[MAXPATHLEN];
FILE *indexf,
     *dataf,
     *oputf;

unsigned long pcntr = 1;	/* pointer counter, also starting pointer */
int doclear = 0;
int gibberish_flag = 0;		/* not implemented! */

void usage(char *calledas)
{
    sdb_version(calledas, NULL);
    fprintf(stderr, "Usage:\n\t%s [-cfh] [-n num] indexfile\n\n", calledas);
    fprintf(stderr, "\t-c\tClear screen between records.\
\n	-f	Force use of an outdated index file.\
\n	-h	Print this us{eless|age} message.\
\n	-n num	Start with record 'num'.\
\n	-V	Print version and exit.\
\n\
\nCommands:\
\n	d	delete the displayed record.\
\n	n	display the next record.\
\n	p	display the previous record.\
\n	q	quit.\
\n	w	append this record to a file\
\n\
\n	filename>	Type the name of the file to append to.\n");
    free(calledas);
    exit(sdb_EUSAGE);
}

void getargs(int ac, char *av[])
{
    int i;
    char *ca;
   
    ca = strdup(av[0]);
    while ((i = getopt(ac, av, "cfhn:")) != EOF)
        switch (i)
        {
            case 'c':
                doclear++;
                break;
            case 'f':
                gibberish_flag++;
                break;
            case 'n':
                if (optarg)
                    pcntr = strtoul(optarg, NULL, 10);
                break;
            case 'V':
                sdb_version(ca, NULL);
                exit(0);
                break;
            case 'h':
            case ':':
            case '?':
            default:
                usage(ca);
                break;
        }
   
    av += optind;
    if (*av)
    {
        strcpy(index_fn, *av);
    }
    else usage(ca);
    free(ca);
}

void getfiles(sdb_header *sdbhp)
{
    struct stat datastat;
    int fec = 0;
    
    indexf = sdb_findindex(index_fn, &fec, gibberish_flag, 0);
    if (!indexf)
    {
        fprintf(stderr, "%s: ", index_fn);
        if (fec)
        {
            fprintf(stderr, "%s", sdb_err_string[fec]);
            exit(fec);
        }
    }
    rewind(indexf);
    fread((sdb_header *)sdbhp, sizeof(sdb_header), 1, indexf);
    if (stat(sdbhp->sdb_filename, &datastat) == -1)
    {
        perror(sdbhp->sdb_filename);
        exit(sdb_ENOENT);
    }
    dataf = fopen(sdbhp->sdb_filename, "r+");
    if (!(dataf))
    {
        perror(sdbhp->sdb_filename);
        exit(sdb_ENOENT);
    }
}

int main(int argc, char *argv[])
{
    sdb_header sdbh;
    char fname[MAXPATHLEN], command = '\0';
    sdb_recptr *rp;
    
    getargs(argc, argv);
    getfiles(&sdbh);
    if (pcntr > sdbh.sdb_total) 
        pcntr = sdbh.sdb_total;
    rp = sdb_pointer(indexf, pcntr);
    if (rp)
    {
        sdb_recalloc(dataf, rp);
        sdb_stripterm(rp, sdbh.sdb_sigch);
    }
    fputs(rp->data, stdout);
    do
    {
        fprintf(stderr, "\n[%lu/%lu] d,n,p,q,w>", pcntr, sdbh.sdb_total);
        
        system("stty raw");
        command = (char)fgetc(stdin);
        system("stty -raw");
        
        switch (command)
        {
            case 'd':
                /* reread the record, because we stripped the term */
                sdb_recfree(rp);
                free(rp);
                rp = NULL;
                rp = sdb_pointer(indexf, pcntr);
                if (rp)
                    sdb_recalloc(dataf, rp);
                sdb_deleterec(dataf, rp->data_os, rp->data_size);
                fprintf(stderr, "\n");
                break;
            case 'n':
                if (doclear) system("clear");
                else fprintf(stderr, "\n\n");
                sdb_recfree(rp);
                free(rp);
                rp = NULL;
                rp = sdb_nextptr(indexf);
                if (rp)
                {
                    sdb_recalloc(dataf, rp);
                    sdb_stripterm(rp, sdbh.sdb_sigch);
                }
                fputs(rp->data, stdout);
                pcntr == sdbh.sdb_total ? pcntr = 1 : pcntr++;
                break;
            case 'p':
                if (doclear) system("clear");
                else fprintf(stderr, "\n\n");
                sdb_recfree(rp);
                free(rp);
                rp = NULL;
                rp = sdb_prevptr(indexf);
                if (rp)
                {
                    sdb_recalloc(dataf, rp);
                    sdb_stripterm(rp, sdbh.sdb_sigch);
                }
                fputs(rp->data, stdout);
                pcntr == 1 ? pcntr = sdbh.sdb_total : pcntr--;
                break;
            case 'w':
                fprintf(stderr, "\nfilename>");
                fgets(fname, MAXPATHLEN, stdin);
                fname[strlen(fname) - 1] = '\0';
                oputf = fopen(fname, "a");
                if (oputf)
                {
                    /* add single-char delimiter, too */
                    sdb_addterm(rp, sdbh.sdb_sigch, 2);
                    fwrite(rp->data, sizeof(char), rp->data_size, oputf);
                    fclose(oputf);
                }
                break;
            default:
                break;
        }
    } while (command != 'q');
    fprintf(stderr, "\n");
    fclose(indexf);
    fclose(dataf);
    return 0;
}
