/*
 * grepstr.c
 *
 * Part of the strings db toolkit.
 *
 * grepstr returns one or more records matching a given string.  It's sort of
 * a grep for the toolkit.
 *
 * For usage, grepstr -h
 *
 * version 0.2.1, first record and all records working (others not),
 *                September 23, 1996, too early in the damned morning.
 * 0.4.1, April 11, 1997.  Library long since revised to support POSIX
 *	regex matching, which gives this tool the same capability, if
 *	compiled in.  This revision gives full functionality; all command
 *	line switches have been enabled.
 * 0.5.0, June 17, 1997.  Minor revisions.
 *
 * Copyright (c) 1996, 1997 Amelia A. Lewis
 * Distributed under GPL.  See COPYING.
 *
 * $Id: grepstr.c 13 2004-03-27 20:21:34Z amyzing $
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <ctype.h>
#include "../lib/strdb.h"
#include "../lib/errdb.h"

char index_fn[MAXPATHLEN], 
     oput_fn[MAXPATHLEN],
     *progname;
FILE *indexf, 
     *dataf, 
     *oputf;

int babble_flag = 0;
int case_flag = 1;
int regex_flag = 0;
int return_type = 0; /* 0 index, 1 offsets, 2 text */
int gibberish_flag = 0;
int whole_record_flag = 0;

unsigned long n_ptrs = 1; /* 0 = return all matches */
unsigned long which_ptr = 1; /* 0 = last */
char *matchstring, 
     matchfield = '\0';

void
usage(char *calledas)
{
    sdb_version(calledas, NULL);
    fprintf(stderr, "Usage:\n\t%s [-FghilortvV] [-f fld] [-m num] [-n num] expr idxfile [ofile]\n\n", calledas);
    fprintf(stderr, "\t-f fld\tSearch field 'fld' (default ambiguous field).\
\n	-F	Force use of outdated index file.\
\n	-g	Global search: report all matches.\
\n	-h	Print this usage message.\
\n	-i	Ignore case in matches.\
\n	-l	Return last match (only).\
\n	-m num	Return a maximum of 'num' matches (default 1).\
\n	-n num	Return match 'num' (only).\
\n	-o	Return offsets (pointers) instead of indices.\
\n	-r	Treat 'expr' as a regular expression (if support compiled in).\
\n	-t	Return text of records instead of indices.\
\n	-v	Be verbose.\
\n	-V	Report version and exit.\
\n	-w	Check entire record, not just specified field.\
\nexpr:		String to match.  May be regex, if support is compiled in.\
\nidxfile:	Name of index file to search.\
\nofile:		Output to named file.\n");
    exit(sdb_EUSAGE);
}

void
getargs(int ac, char *av[])
{
    int i;
    
    if (strrchr(av[0], '/'))
        progname = strdup(strrchr(av[0], '/'));
    else
       progname = strdup(av[0]);
    while ((i = getopt(ac, av, "f:Fghilm:n:ortvVw")) != EOF)
        switch (i)
        {
            case 'f': /* field */
                if (isgraph(*optarg))
                    matchfield = *optarg;
                break;
            case 'F': /* use outdated index */
                gibberish_flag++;
                break;
            case 'g': /* global, return all */
                n_ptrs = 0;
                which_ptr = 1;
                break;
            case 'i': /* ignore case */
                case_flag--;
                break;
            case 'l': /* return last */
                which_ptr = 0;
                break;
            case 'm': /* return no more than num records */
                if (*optarg)
                    n_ptrs = strtoul(optarg, NULL, 10);
                else
                    usage(progname);
                break;
            case 'n': /* return matching record number num */
                if (*optarg)
                    which_ptr = strtoul(optarg, NULL, 10);
                else
                    usage(progname);
                break;
            case 't': /* return the text of the pointers, not the indices */
                return_type = 2;
                break;
            case 'o': /* return the pointers (offsets), not the indices */
                return_type = 1;
                break;
            case 'r': /* match is a regular expression */
                regex_flag++;
                break;
            case 'v':
                babble_flag++;
                break;
            case 'V':
                sdb_version(progname, NULL);
                exit(0);
                break;
            case 'w':
                whole_record_flag++;
                break;
            case 'h':
            case ':':
            case '?':
            default:
                usage(progname);
                break;
        }

    if (n_ptrs && which_ptr)
        n_ptrs = (n_ptrs > which_ptr) ? n_ptrs : which_ptr;
    if (whole_record_flag)
        matchfield = '\0';
    av += optind;
    if (*av)
        strcpy(matchstring, *av);
    else
        usage(progname);
    if (*++av)
        strcpy(index_fn, *av);
    else usage(progname);
    if (*++av)
        strcpy(oput_fn, *av);
    else strcpy(oput_fn, "-");
}

void 
getfiles(sdb_header *sdbhp)
{
    int filerr = 0;

    indexf = sdb_findindex(index_fn, &filerr, gibberish_flag, 0);
    if (!indexf)
    {
        fprintf(stderr, "%s: %s: %s", progname, index_fn, sdb_err_string[filerr]);
        exit(filerr);
    }
    rewind(indexf);
    if (!fread((sdb_header *)sdbhp, sizeof(sdb_header), 1, indexf))
    {
        fprintf(stderr, "%s: %s", progname, sdb_err_string[sdb_EFREAD]);
        exit(sdb_EFREAD);
    }

    dataf = fopen(sdbhp->sdb_filename, "r");
    if (!(dataf))
    {
        perror(sdbhp->sdb_filename);
        exit(sdb_ENOENT);
    }

    if (strcmp(oput_fn, "-") == 0)
        oputf = stdout;
    else 
        oputf = fopen(oput_fn, "w+");
    if (!(oputf))
    {
        perror(oput_fn);
        exit(sdb_ENOENT);
    }
}

int
main(int argc, char *argv[])
{
    off_t *oarray;
    unsigned long n_ret, i;
    sdb_header sdbh;
    sdb_recptr *srp = NULL;

    matchstring = (char *)calloc(MAXLINELEN, sizeof(char));
    if (!(matchstring))
    {
        fprintf(stderr, "%s: %s", argv[0], sdb_err_string[sdb_EMEMFAULT]);
        exit(sdb_EMEMFAULT);
    }
    getargs(argc, argv);
    getfiles(&sdbh);

    /* this sets the maximum number to return to the minimum of n_ptrs or
       the total number of pointers in the file */
    n_ret = (n_ptrs > sdbh.sdb_total) ? sdbh.sdb_total : n_ptrs;
    /* but if the number set was 0, that's the key that we want as
       many as possible, so reset to total number */
    if (!n_ret) n_ret = sdbh.sdb_total;
    
    /* and make an array that can hold that many pointers */
    oarray = (off_t *)calloc(n_ret, sizeof(off_t));

    if (!(oarray))
    {
        fprintf(stderr, "%s: %s", progname, sdb_err_string[sdb_EMEMFAULT]);
        free(matchstring);
        exit(sdb_EMEMFAULT);
    }
    
    /* Now call the library function to return the pointers */
    n_ret = sdb_grep(dataf, indexf, matchstring, matchfield, oarray, 
                     n_ret, case_flag, return_type, regex_flag,
                     whole_record_flag, which_ptr);

    /* output routines, vary based on babble_flag and type flags */

    if (!which_ptr)
    {
        oarray[0] = oarray[n_ret - 1];
        n_ret = 1;
    }
    else if (which_ptr != 1)
    {
        if (which_ptr <= n_ret)
        {
            oarray[0] = oarray[which_ptr - 1];
            n_ret = 1;
        }
        else
            n_ret = 0;
    }
    if (babble_flag)
        fprintf(stderr, "%lu match%s.\n", n_ret, (n_ret != 1) ? "es" : "");
    for (i = 0; i < n_ret ; i++)
    {
        if (return_type == 0)
        {
            if (babble_flag)
                fprintf(oputf, "Index: %ld\n", oarray[i]);
            else
                fprintf(oputf, "%ld ", oarray[i]);
        }
        else if (return_type == 1)
        {
            if (babble_flag)
                fprintf(oputf, "Offset: %lu\n", oarray[i]);
            else
                fprintf(oputf, "%lu ", oarray[i]);
        }
        else
        {
            srp = sdb_pointer(indexf, oarray[i]);
            if (srp)
            {
                if (sdb_recalloc(dataf, srp))
                {
                    fprintf(oputf, "%s", srp->data);
                    sdb_recfree(srp);
                }
                free(srp);
                srp = NULL;
            }
        }
    }
    if (!babble_flag && (return_type < 2))
        fprintf(oputf, "\n");

    free(matchstring);
    free(oarray);
    fclose(indexf);
    fclose(dataf);
    fclose(oputf);
    return 0;
}

