/*
 * idxstr.c
 *
 * Part of the strings db toolkit.
 *
 * idxstr creates the index file used by other utilities.
 *
 * For usage instructions, use idxstr -h (or idxstr -invalidcharacter)
 * 
 * version 0.2.0, September 22, 1996 (== 0.2.1)
 * November 23, 1996: added reverse function for sorting.
 * 0.4.0, March 24 1997: considerably cleaned up.
 * 0.4.2, June 15 1997: added babbling.
 *
 * Copyright (c) 1996 Amelia A. Lewis
 * Distributed under GPL.  See COPYING.
 *
 * $Id: idxstr.c 13 2004-03-27 20:21:34Z amyzing $
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <ctype.h>
#include <stdlib.h>
#include <time.h> /* sigh */
#include "../lib/strdb.h"
#include "../lib/sortdb.h"
#include "../lib/errdb.h"

char data_fn[MAXPATHLEN], index_fn[MAXPATHLEN];
char *progname;
FILE *dataf, *indexf;

int babble_flag = 0;
int sort_flag = 0;
int update_flag = 0;

unsigned char sort_flags;
unsigned char app_flags = 0;	/* not implemented */

char sort_field = '\0';
char data_sc;

/*
 * usage
 *	Prints a usage statement.  Exits with return 1.
 */
void
usage(char *calledas)
{
    sdb_version(calledas, NULL);
    fprintf(stderr, "Usage:\n\t%s [-ahirRsvw] [-A hex] [-c sch] [-f fld] datafile [.ext|idxfile]\
\n\
\n	-A hex	Set application flags nn (example: 20 = flag 5)\
\n	-a	Sort on alphanumeric characters.\
\n	-c sch	Use signal character 'sch'.\
\n	-f fld	Sort on field specified by character 'fld'.\
\n	-h	Print this usage message.\
\n	-i	Sort, ignoring case.\
\n	-n	Sort first digit-string as an (unsigned) integer.\
\n	-R	Randomize order.\
\n	-r	Sort in reverse order.\
\n	-s	Sort (on the ambiguous field, by default).\
\n	-u	Update, using flag values in the existing file.\
\n	-v	Be verbose, report statistics.\
\n	-V	Report version and exit.\
\n	-w	Sort, ignoring whitespace.\
\n\
\ndatafile:	Strings file to index.\
\n.ext:		Extension to give index file.\
\nidxfile:	Name to give index file.\
\n\
\nNote: any of the sorting options (afiRw) also turn on sorting\
\n	(-s need not be specified).\
\n	Also, -a implies -w.\n", calledas);
    exit(sdb_EUSAGE);
}

/*
 * getargs
 *	Parses the command line, setting flags.
 */
void
getargs(int ac, char *av[])
{
    int c;
    char isuff[MAXPATHLEN];
    
    if (strrchr(av[0], '/'))
        progname = strdup(strrchr(av[0], '/'));
    else
        progname = strdup(av[0]);
    data_sc = sdb_DEFSC;
    sort_flags = sdb_SORT_CASE;
    while ((c = getopt(ac, av, "A:ac:f:hinrRsuvVw")) != EOF)
        switch (c)
        {
            case 'A':
                if (*optarg)
                    app_flags = strtoul(optarg, NULL, 16);
                break;
            case 'a':
                sort_flags |= sdb_SORT_ALNUM;
                sort_flag++;
                break;
            case 'c':
                if (isgraph(*optarg))
                    data_sc = *optarg;
                break;
            case 'f':
                if (isgraph(*optarg))
                    sort_field = *optarg;
                sort_flag++;
                break;
            case 'i':
                sort_flags &= ~(sdb_SORT_CASE);
                sort_flag++;
                break;
            case 'n':
                sort_flags |= sdb_SORT_NUMBER;
                sort_flag++;
                break;
            case 'R':
                sort_flag--;
                break;
            case 'r':
                sort_flags |= sdb_SORT_REVERSE;
                sort_flag++;
                break;
            case 's':
                sort_flag++;
                break;
            case 'u':
                update_flag++;
                break;
            case 'v':
                babble_flag++;
                break;
            case 'V':
                sdb_version(progname, NULL);
                exit(sdb_EUSAGE);
                break;
            case 'w':
                sort_flags |= sdb_SORT_IWHITE;
                sort_flag++;
                break;
            case 'h':
            case ':':
            case '?':
            default:
                usage(progname);
                break;
        }

    if (sort_flag && !sort_field)
        sort_field = data_sc;
    if (sort_flags & sdb_SORT_NUMBER)
        sort_flags &= (sdb_SORT_NUMBER | sdb_SORT_REVERSE);	/* insurance to avoid thrash */
    av += optind;
    if (*av)
        strcpy(data_fn, *av);
    else usage(progname);
    if (data_fn[0] == '-')
    {
        fprintf(stderr, "Input from stdin not allowed.\n");
        usage(progname);
    }
    if (*++av)
    {
        strcpy(isuff, *av);
        if ((isuff[0] == '.') && (isuff[1] != '/'))
        {
            strcpy(index_fn, data_fn);
            strcat(index_fn, isuff);
        }
        else strcpy(index_fn, isuff);
    }
    else
    {
        strcpy(index_fn, data_fn);
        if (sort_field && (sort_field != data_sc))
        {
            sprintf(isuff, ".%cdx", sort_field);
            strcat(index_fn, isuff);
        }
        else strcat(index_fn, ".index");
    }
    if (babble_flag)
    {
    	fprintf(stderr, "Creating index %s for data file %s\n", index_fn, data_fn);
    	fprintf(stderr, "Flags: ");
    	if (sort_flag)
    	{
    	    fprintf(stderr, "Sort: ");
    	    if (sort_flags & sdb_SORT_IWHITE)
    	        fprintf(stderr, "ignore-whitespace ");
    	    else if (sort_flags & sdb_SORT_ALNUM)
    	        fprintf(stderr, "alphanumeric ");
    	    if (sort_flags & sdb_SORT_CASE)
    	        fprintf(stderr, "case-sensitive ");
    	    if (sort_flags & sdb_SORT_REVERSE)
    	        fprintf(stderr, "reverse ");
    	    if (sort_flags & sdb_SORT_NUMBER)
    	        fprintf(stderr, "numeric ");
    	}
    	if (update_flag)
    	    fprintf(stderr, "Updating ");
    	if (app_flags)
    	{
    	    fprintf(stderr, "Application: ");
    	    if (app_flags & 0x80)
    	        fprintf(stderr, "7");
    	    if (app_flags & 0x40)
    	        fprintf(stderr, "6");
    	    if (app_flags & 0x20)
    	        fprintf(stderr, "5");
    	    if (app_flags & 0x10)
    	        fprintf(stderr, "4");
    	    if (app_flags & 0x08)
    	        fprintf(stderr, "3");
    	    if (app_flags & 0x04)
    	        fprintf(stderr, "2");
    	    if (app_flags & 0x02)
    	        fprintf(stderr, "1");
    	    if (app_flags & 0x01)
    	        fprintf(stderr, "0");
    	}
    	fprintf(stderr, "\n");
    	if (sort_flag)
    	    fprintf(stderr, "Sorting on field: %c\n", sort_field);
    	fprintf(stderr, "Signal character: %c\n", data_sc);
    }
}

/*
 * getfiles
 *	Opens the files.  If update is set, reads flags from the old file.
 */
void
getfiles(void)
{
    sdb_header sdbh;
    
    dataf = fopen(data_fn, "r");
    if (!(dataf))
    {
        perror(data_fn);
        exit(sdb_ENOENT);
    }
    if (update_flag)
    {
        indexf = fopen(index_fn, "r");
        if (indexf)
        {
            rewind(indexf);
            if (fread((sdb_header *)&sdbh, sizeof(sdb_header), 1, indexf))
            {
                if (data_sc != sdb_DEFSC)
                    data_sc = sdbh.sdb_sigch;
                sort_flags = sdbh.sdb_sortflags;
                app_flags = sdbh.sdb_appflags;
                sort_flag = sdbh.sdb_sorton;
                sort_field = sdbh.sdb_sorton;
            }
            else
                fprintf(stderr, "Glitch\n");
            fclose(indexf);
        }
        else
            fprintf(stderr, "Index file %s not found.\nUsing default flags.\n", index_fn);
    }
    indexf = fopen(index_fn, "w+");
    if (!(indexf))
    {
        perror(index_fn);
        exit(sdb_ENOENT);
    }
}

/*
 * init_header
 *	Initializes the sdb_header structure.
 */
void
init_header(sdb_header *sdbhp)
{
    struct stat datastat;
    char *tempfn;
    
    if (babble_flag)
        fprintf(stderr, "Writing header\n");
    sdbhp->sdb_major = sdb_VMAJOR;
    sdbhp->sdb_minor = sdb_VMINOR;
    sdbhp->sdb_patch = sdb_VPATCH;
    sdbhp->sdb_sigch = data_sc;
    sdbhp->sdb_sortflags = sort_flags;
    sdbhp->sdb_reserved_nonsense = 0;
    sdbhp->sdb_appflags = app_flags;
    if (!sort_flag) 
        sdbhp->sdb_sorton = '\0';
    else
    {
        if (sort_field != data_sc)
            sdbhp->sdb_sorton = sort_field;
        else
            sdbhp->sdb_sorton = data_sc;
    }
    strcpy(sdbhp->sdb_magic, sdb_MAGIC);
    sdbhp->sdb_total = 0;
    stat(data_fn, &datastat);
    sdbhp->sdb_mtime = datastat.st_mtime;
    tempfn = realpath(data_fn, sdbhp->sdb_filename);
    if (!tempfn)
        strcpy(sdbhp->sdb_filename, data_fn);
}

/*
 * get_next_record
 *	Reads the next record, allocating and initializing an sdb_recptr
 * structure.
 */
sdb_recptr *
get_next_record(void)
{
    sdb_recptr *thisrec;
    char linein[MAXLINELEN], *lptr, delim;
    int has_printable = 0;
    
    if (feof(dataf)) return NULL;
    thisrec = (sdb_recptr *)malloc(sizeof(sdb_recptr));
    thisrec->next = NULL;
    thisrec->data = NULL;
    thisrec->data_os = ftell(dataf);
    do
    {
        fgets(linein, MAXLINELEN, dataf);
        delim = sdb_delim(linein, data_sc);
        if (!(has_printable))
            if (!(delim))
                for ( lptr = linein; *lptr ; lptr++ )
                    if (isgraph(*lptr))
                    {
                        has_printable++;
                        break;
                    }
    } while ((delim != 1) && (delim != '\b') && !(feof(dataf)));
    if (!(has_printable) || (delim == '\b'))
    {
        thisrec->data_os = -1;
        thisrec->data_size = 0;
    }
    else
    	thisrec->data_size = ftell(dataf) - thisrec->data_os;
    return thisrec;
}

/*
 * write_pointers
 *	Writes pointers to the index file (offset and size)
 */
void
write_pointers(sdb_recptr *pstart, sdb_header *sdbhp)
{
    sdb_recptr *pcur;

    if (babble_flag)
        fprintf(stderr, "Writing pointers ");    
    while (pstart)
    {
        if (babble_flag)
            fprintf(stderr, ".");
        pcur = pstart;
        pstart = (sdb_recptr *)pcur->next;
        if (!fwrite((off_t *)&pcur->data_os, sizeof(off_t), 1, indexf))
        {
            fprintf(stderr, "strdb: %s", sdb_err_string[sdb_EFWRITE]);
            exit(sdb_EFWRITE);
        }
        if (!fwrite((unsigned short *)&pcur->data_size, sizeof(unsigned short), 1, indexf))
        {
            fprintf(stderr, "strdb: %s", sdb_err_string[sdb_EFWRITE]);
            exit(sdb_EFWRITE);
        }
        free(pcur);
    }
    if (babble_flag)
        fprintf(stderr, " done\n");
}

/*
 * randomize_pointers
 *	A really stupid function, but it's traditional.
 */
void
randomize_pointers(sdb_recptr *pstart, unsigned long np)
{
    unsigned long cnt = 0, l;
    long ranum;
    sdb_recptr *pcur, ptemp, *ocur;
    
    /* you know, randomizing pointers is fairly pointless */
    if (babble_flag)
        fprintf(stderr, "Randomizing ");
    if (!pstart)
    {
        if (babble_flag)
            fprintf(stderr, "done (no records)\n");
            return;
    }
    pcur = pstart;
    while (pcur->next)
        pcur = pcur->next;
    pcur->next = pstart;	/* circular list.  whoopee. */
    srandom((int) (time((time_t *) NULL) + getpid()));
    pcur = ocur = pstart;
    while (cnt < np)
    {
        if (babble_flag)
            fprintf(stderr, ".");
        ranum = 1 + (long)(np * random() / (RAND_MAX + 1.0));
        for (l = 0; l < ranum ; l++)
            pcur = pcur->next;
        ranum = 1 + (long)(np * random() / (RAND_MAX + 1.0));
        for (l = 0; l < ranum ; l++)
            ocur = ocur->next;
        ptemp.data_os = pcur->data_os; /* flip */
        pcur->data_os = ocur->data_os; /* flap */
        ocur->data_os = ptemp.data_os; /* flop!*/
        cnt += 2;
    }
    pcur = pstart;
    while (pcur->next != pstart)
        pcur = pcur->next;
    pcur->next = NULL;	/* break the chain */
    if (babble_flag)
        fprintf(stderr, "done\n");
}

/*
 * main
 *	If you need a comment for main, I suggest you try K&R
 */
int
main(int argc, char *argv[])
{
    sdb_header sdbh;
    sdb_recptr *currec = NULL, *firstrec = NULL, *prevrec;
    char comstr[MAXPATHLEN + 12];
    
    getargs(argc, argv);
    getfiles();
    init_header(&sdbh);
    if (babble_flag)
        fprintf(stderr, "Parsing records ");
    do
    {
        if (babble_flag)
            fprintf(stderr, ".");
        prevrec = currec;
        currec = get_next_record();
        if (currec->data_os == -1)
        {
            free(currec);
            currec = prevrec;
        }
        else
        {
            sdbh.sdb_total++;
            if (prevrec)
                prevrec->next = currec;
            else
                firstrec = currec;
        }
    } while (!feof(dataf));
    if (babble_flag)
        fprintf(stderr, " done\n");
    if (sort_flag > 0)
        firstrec = sdb_llmsort(firstrec, sdb_comparefield, sort_flags, 
                               data_sc, ((sort_field == data_sc) ? '\0' : sort_field), 
                               dataf, babble_flag);
    if (sort_flag < 0) randomize_pointers(firstrec, sdbh.sdb_total);
    
    fclose(dataf);
    fwrite((sdb_header *)&sdbh, sizeof(sdb_header), 1, indexf);
    write_pointers(firstrec, &sdbh);
    fclose(indexf);
    if (babble_flag)
    {
        sprintf(comstr, "lsstr -Av %s", index_fn);
        system(comstr);
    }
    return 0;
}
